from rest_framework import status
from rest_framework.reverse import reverse
from .base import BaseTest


class PaymentMethodsViewTest(BaseTest):
    def test_introspection_authenticated(self):
        self.login(is_staff=True)
        url = reverse('api-checkout-payment-methods')
        resp = self.client.get(url)
        self.assertEquals(resp.status_code, status.HTTP_200_OK)
        self.assertEquals(resp.data, {
            'credit-card': {
                'type': 'nested object',
                'required': False,
                'read_only': False,
                'label': 'Credit-card',
                'children': {
                    'enabled': {
                        'type': 'boolean',
                        'required': False,
                        'read_only': False,
                        'label': 'Enabled',
                    },
                    'pay_balance': {
                        'type': 'boolean',
                        'required': False,
                        'read_only': False,
                        'label': 'Pay balance',
                    },
                    'amount': {
                        'type': 'decimal',
                        'required': False,
                        'read_only': False,
                        'label': 'Amount',
                    },
                    'reference': {
                        'type': 'string',
                        'required': False,
                        'read_only': False,
                        'label': 'Reference',
                        'max_length': 128,
                    },
                },
            },
            'cash': {
                'type': 'nested object',
                'required': False,
                'read_only': False,
                'label': 'Cash',
                'children': {
                    'enabled': {
                        'type': 'boolean',
                        'required': False,
                        'read_only': False,
                        'label': 'Enabled',
                    },
                    'pay_balance': {
                        'type': 'boolean',
                        'required': False,
                        'read_only': False,
                        'label': 'Pay balance',
                    },
                    'amount': {
                        'type': 'decimal',
                        'required': False,
                        'read_only': False,
                        'label': 'Amount',
                    },
                    'reference': {
                        'type': 'string',
                        'required': False,
                        'read_only': False,
                        'label': 'Reference',
                        'max_length': 128,
                    },
                },
            },
        })


    def test_introspection_anonymous(self):
        url = reverse('api-checkout-payment-methods')
        resp = self.client.get(url)
        self.assertEquals(resp.status_code, status.HTTP_200_OK)
        self.assertEquals(resp.data, {
            'credit-card': {
                'type': 'nested object',
                'required': False,
                'read_only': False,
                'label': 'Credit-card',
                'children': {
                    'enabled': {
                        'type': 'boolean',
                        'required': False,
                        'read_only': False,
                        'label': 'Enabled',
                    },
                    'pay_balance': {
                        'type': 'boolean',
                        'required': False,
                        'read_only': False,
                        'label': 'Pay balance',
                    },
                    'amount': {
                        'type': 'decimal',
                        'required': False,
                        'read_only': False,
                        'label': 'Amount',
                    },
                    'reference': {
                        'type': 'string',
                        'required': False,
                        'read_only': False,
                        'label': 'Reference',
                        'max_length': 128,
                    },
                },
            },
        })



PENDING = 'Pending'
DECLINED = 'Declined'
COMPLETE = 'Complete'


class PaymentStatus(object):
    def __init__(self, amount):
        self.amount = amount

    @property
    def status(self):
        raise NotImplementedError('Subclass must implement status property')

    def get_required_action(self):
        raise NotImplementedError('Subclass does not implement get_required_action()')


class Complete(PaymentStatus):
    status = COMPLETE


class Declined(PaymentStatus):
    status = DECLINED


class FormPostRequired(PaymentStatus):
    status = PENDING

    def __init__(self, amount, name, url, method='POST', fields=[]):
        super().__init__(amount)
        self.form_data = {
            'type': 'form',
            'name': name,
            'url': url,
            'method': method,
            'fields': fields,
        }

    def get_required_action(self):
        return self.form_data



from django.dispatch import receiver
from .email import OrderMessageSender
from .settings import ORDER_PLACED_COMMUNICATION_TYPE_CODE
from .signals import order_payment_authorized


@receiver(order_payment_authorized)
def send_order_confirmation_message(sender, order, request, **kwargs):
    OrderMessageSender(request).send_confirmation_message(order, ORDER_PLACED_COMMUNICATION_TYPE_CODE)



from django.shortcuts import get_object_or_404
from rest_framework import generics, status
from rest_framework.response import Response
from oscar.core.loading import get_model
from .serializers import (
    CheckoutSerializer,
    OrderSerializer,
    PaymentMethodsSerializer,
    PaymentStateSerializer
)
from .signals import order_placed
from . import utils

Order = get_model('order', 'Order')

CHECKOUT_ORDER_ID = 'checkout_order_id'


class PaymentMethodsView(generics.GenericAPIView):
    serializer_class = PaymentMethodsSerializer

    def get(self, request):
        serializer = self.get_serializer()
        data = self.metadata_class().get_serializer_info(serializer)
        return Response(data)


class CheckoutView(generics.GenericAPIView):
    """
    Checkout and use a WFRS account as payment.

    POST(basket, shipping_address, wfrs_source_account,
         [total, shipping_method_code, shipping_charge, billing_address]):
    {
        "basket": "/api/baskets/1/",
        "guest_email": "foo@example.com",
        "total": "100.0",
        "shipping_charge": {
            "currency": "EUR",
            "excl_tax": "10.0",
            "tax": "0.6"
        },
        "shipping_method_code": "no-shipping-required",
        "shipping_address": {
            "country": "/api/countries/NL/",
            "first_name": "Henk",
            "last_name": "Van den Heuvel",
            "line1": "Roemerlaan 44",
            "line2": "",
            "line3": "",
            "line4": "Kroekingen",
            "notes": "Niet STUK MAKEN OK!!!!",
            "phone_number": "+31 26 370 4887",
            "postcode": "7777KK",
            "state": "Gerendrecht",
            "title": "Mr"
        },
        "payment": {
            "cash": {
                "amount": "100.00"
            }
        }
    }

    Returns the order object.
    """
    serializer_class = CheckoutSerializer

    def post(self, request, format=None):
        # Wipe out any previous state data
        utils.clear_payment_method_states(request)

        # Validate the input
        c_ser = self.get_serializer(data=request.data)
        if not c_ser.is_valid():
            return Response(c_ser.errors, status.HTTP_406_NOT_ACCEPTABLE)

        # Freeze basket
        basket = c_ser.validated_data.get('basket')
        basket.freeze()

        # Save Order
        order = c_ser.save()
        request.session[CHECKOUT_ORDER_ID] = order.id

        # Save payment steps into session for processing
        states = self._record_payments(
            request=request,
            order=order,
            methods=c_ser.fields['payment'].methods,
            data=c_ser.validated_data['payment'])
        utils.set_payment_method_states(order, request, states)

        # Return order data
        order_placed.send(sender=self, order=order, user=request.user, request=request)
        o_ser = OrderSerializer(order, context={ 'request': request })
        return Response(o_ser.data)


    def _record_payments(self, request, order, methods, data):
        order_balance = [order.total_incl_tax]
        states = {}

        def record(code, method_data):
            method = methods[code]
            state = method.record_payment(request, order, **method_data)
            order_balance[0] = order_balance[0] - state.amount
            return state

        # Loop through each method with a specified amount to charge
        data_amount_specified = { k: v for k, v in data.items() if not v['pay_balance'] }
        for code, method_data in data_amount_specified.items():
            states[code] = record(code, method_data)

        # Change the remainder, not covered by the above methods, to the method marked with `pay_balance`
        data_pay_balance = { k: v for k, v in data.items() if v['pay_balance'] }
        for code, method_data in data_pay_balance.items():
            method_data['amount'] = order_balance[0]
            states[code] = record(code, method_data)

        return states


class PaymentStatesView(generics.GenericAPIView):
    def get(self, request, pk=None):
        # We don't really use the provided pk. It's just there to be compatible with oscarapi
        if pk and int(pk) != request.session.get(CHECKOUT_ORDER_ID):
            return Response(status=status.HTTP_404_NOT_FOUND)

        # Fetch the order object from the session, but only if it's pending
        pk = request.session.get(CHECKOUT_ORDER_ID)
        order = get_object_or_404(Order, pk=pk)

        # Return order status and payment states
        states = utils.list_payment_method_states(request)
        state_data = {}
        for code, state in states.items():
            ser = PaymentStateSerializer(instance=state)
            state_data[code] = ser.data

        return Response({
            'order_status': order.status,
            'payment_method_states': state_data if any(state_data) else None
        })



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


def overridable(name, default=None, required=False, cast=None):
    if required:
        if not hasattr(settings, name) or not getattr(settings, name):
            raise ImproperlyConfigured("%s must be defined in Django settings" % name)
    value = getattr(settings, name, default)
    if cast:
        value = cast(value)
    return value


API_ENABLED_PAYMENT_METHODS = overridable('API_ENABLED_PAYMENT_METHODS', [
    {
        'method': 'oscarapicheckout.methods.Cash',
        'permission': 'oscarapicheckout.permissions.StaffOnly',
    }
])
API_MAX_PAYMENT_METHODS = overridable('API_MAX_PAYMENT_METHODS', 0)

ORDER_STATUS_PENDING = overridable('ORDER_STATUS_PENDING', 'Pending')
ORDER_STATUS_PAYMENT_DECLINED = overridable('ORDER_STATUS_PAYMENT_DECLINED', 'Payment Declined')
ORDER_STATUS_AUTHORIZED = overridable('ORDER_STATUS_AUTHORIZED', 'Authorized')
ORDER_STATUS_SHIPPED = overridable('ORDER_STATUS_SHIPPED', 'Shipped')
ORDER_STATUS_CANCELED = overridable('ORDER_STATUS_CANCELED', 'Canceled')

ORDER_PLACED_COMMUNICATION_TYPE_CODE = overridable('ORDER_PLACED_COMMUNICATION_TYPE_CODE', 'ORDER_PLACED')

ORDER_OWNERSHIP_CALCULATOR = overridable('ORDER_OWNERSHIP_CALCULATOR', 'oscarapicheckout.utils.get_order_ownership')



from django.apps import AppConfig


class Config(AppConfig):
    name = 'oscarapicheckout'
    label = 'oscarapicheckout'

    def ready(self):
        # Register signal handlers
        from . import handlers  # NOQA



from django.conf.urls import url
from django.views.decorators.cache import never_cache
from oscar.core.application import Application
from .views import PaymentMethodsView, CheckoutView, PaymentStatesView


class APICheckoutApplication(Application):
    def get_urls(self):
        view_methods = never_cache(PaymentMethodsView.as_view())
        view_states = never_cache(PaymentStatesView.as_view())
        view_checkout = never_cache(CheckoutView.as_view())
        urlpatterns = [
            url(r'^checkout/payment-methods/$', view_methods, name='api-checkout-payment-methods'),
            url(r'^checkout/payment-states/$', view_states, name='api-payment'),
            url(r'^checkout/payment-states/(?P<pk>\d+)/$', view_states, name='api-payment'),
            url(r'^checkout/$', view_checkout, name='api-checkout'),
        ]
        return self.post_process_urls(urlpatterns)


application = APICheckoutApplication()



from decimal import Decimal
from django.db import transaction
from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers
from oscar.core.loading import get_model
from . import states

PaymentEventType = get_model('order', 'PaymentEventType')
PaymentEvent = get_model('order', 'PaymentEvent')
PaymentEventQuantity = get_model('order', 'PaymentEventQuantity')
SourceType = get_model('payment', 'SourceType')
Source = get_model('payment', 'Source')
Transaction = get_model('payment', 'Transaction')


class PaymentMethodSerializer(serializers.Serializer):
    enabled = serializers.BooleanField(default=False)
    pay_balance = serializers.BooleanField(default=True)
    amount = serializers.DecimalField(
        decimal_places=2,
        max_digits=12,
        required=False)
    reference = serializers.CharField(
        max_length=128,
        default='')

    def validate(self, data):
        if not data['enabled']:
            return data

        if data['pay_balance']:
            data.pop('amount', None)
        elif 'amount' not in data or not data['amount'] or data['amount'] <= Decimal('0.00'):
            raise serializers.ValidationError(_("Amount must be greater then 0.00 or pay_balance must be enabled."))

        return data


class PaymentMethod(object):
    name = _('Abstract Payment Method')
    code = 'abstract-payment-method'
    serializer_class = PaymentMethodSerializer

    def _make_payment_event(self, type_name, order, amount, reference=''):
        etype, created = PaymentEventType.objects.get_or_create(name=type_name)
        event = PaymentEvent()
        event.order = order
        event.amount = amount
        event.reference = reference if reference else ''
        event.event_type = etype
        event.save()
        return event

    def make_authorize_event(self, *args, **kwargs):
        return self._make_payment_event(Transaction.AUTHORISE, *args, **kwargs)

    def make_debit_event(self, *args, **kwargs):
        return self._make_payment_event(Transaction.DEBIT, *args, **kwargs)

    def make_refund_event(self, *args, **kwargs):
        return self._make_payment_event(Transaction.REFUND, *args, **kwargs)

    def make_event_quantity(self, event, line, quantity):
        return PaymentEventQuantity.objects.create(event=event, line=line, quantity=quantity)

    def get_source(self, order, reference=''):
        stype, created = SourceType.objects.get_or_create(name=self.name)
        source, created = Source.objects.get_or_create(order=order, source_type=stype)
        source.currency = order.currency
        source.reference = reference
        source.save()
        return source

    @transaction.atomic()
    def record_payment(self, request, order, amount=None, reference='', **kwargs):
        if not amount and amount != Decimal('0.00'):
            raise RuntimeError('Amount must be specified')
        return self._record_payment(request, order, amount, reference, **kwargs)

    def _record_payment(self, request, order, amount, reference, **kwargs):
        raise NotImplementedError('Subclass must implement _record_payment(request, order, amount, reference, **kwargs) method.')


class Cash(PaymentMethod):
    """
    Cash payments are an example of how to implement a payment method plug-in. It
    doesn't do anything more than record a transaction and payment source.
    """
    name = _('Cash')
    code = 'cash'

    def _record_payment(self, request, order, amount, reference, **kwargs):
        source = self.get_source(order, reference)

        amount_to_allocate = amount - source.amount_allocated
        source.allocate(amount_to_allocate, reference)

        amount_to_debit = amount - source.amount_debited
        source.debit(amount_to_debit, reference)

        event = self.make_debit_event(order, amount_to_debit, reference)
        for line in order.lines.all():
            self.make_event_quantity(event, line, line.quantity)

        return states.Complete(source.amount_debited)



from oscar.core.loading import get_class

OrderPlacementMixin = get_class('checkout.mixins', 'OrderPlacementMixin')


class OrderMessageSender(OrderPlacementMixin):
    def __init__(self, request):
        self.request = request
